package core

import (
	"fmt"

	"git.urbach.dev/cli/q/src/asm"
	"git.urbach.dev/cli/q/src/errors"
	"git.urbach.dev/cli/q/src/eval"
	"git.urbach.dev/cli/q/src/expression"
	"git.urbach.dev/cli/q/src/types"
)

// EvaluateArray evaluates an array access.
func (f *Function) EvaluateArray(expr *expression.Expression) (*eval.Memory, error) {
	name := expr.Children[0].Token.Text(f.File.Bytes)
	base := f.VariableByName(name)

	if base == nil {
		return nil, errors.New(&errors.UnknownIdentifier{Name: name}, f.File, expr.Children[0].Token.Position)
	}

	defer f.UseVariable(base)

	value := &eval.Memory{
		Memory: asm.Memory{
			Base:           base.Value.Register,
			Offset:         0,
			OffsetRegister: -1,
			Length:         0,
		},
	}

	switch baseType := base.Value.Typ.(type) {
	case *types.Array:
		value.Typ = baseType.Of
		value.Memory.Length = byte(baseType.Of.Size())
	case *types.Pointer:
		value.Typ = baseType.To
		value.Memory.Length = byte(baseType.To.Size())
	default:
		return nil, errors.New(&errors.TypeMismatch{Encountered: base.Value.Typ.Name(), Expected: types.AnyPointer.Name()}, f.File, expr.Token.Position)
	}

	if len(expr.Children) == 1 {
		return value, nil
	}

	indexExpr := expr.Children[1]
	index, err := f.Evaluate(indexExpr)

	if err != nil {
		return nil, err
	}

	if !types.Is(index.Type(), types.AnyInt) {
		return nil, errors.New(&errors.TypeMismatch{Encountered: index.Type().Name(), Expected: types.AnyInt.Name()}, f.File, indexExpr.Token.Position)
	}

	switch index := index.(type) {
	case *eval.Number:
		value.Memory.Offset = int8(index.Number)
	case *eval.Register:
		value.Memory.OffsetRegister = index.Register
	default:
		panic(fmt.Errorf("%s: not implemented: %v", f.UniqueName, index))
	}

	return value, nil
}
